import os
import re
import json
import string
import difflib
import urllib.parse

from bs4 import BeautifulSoup
from html2text import HTML2Text
from collections import defaultdict
from phonenumbers import PhoneNumberMatcher, PhoneNumberFormat, format_number, parse

from profilescout.link.utils import to_key, is_url, to_abs_path


PATTERNS = {'unwanted_tag__has_placeholder': r'(<\/?(?:b|i|strong|em|blockquote|h[1-6])\b[^>]*>)',
            'md_link': r'\[(.*?)\]\((.+?)\)',
            'email': r'(?:[a-z0-9_\.\+-]+)@(?:[\da-z\.-]+)\.(?:[a-z\.]{2,6})',
            'different_line': r'^\+*([^+]*\w+.*)$',
            'label_field': r'^\w.*:$',
            'label_field_with_value': r'^(\w.*?):(.*\w.*)$',
            'repetetive_punct': f'[ {re.escape(string.punctuation)}]+',
            'repeating_whitespace': r'(?:(\ )+)|(?:(\t)+)|(?:(\n)+)|(?:(\r)+)|(?:(\f)+)',
            'number': r'((?:\(?(?:00|\+)(?:[1-4]\d\d|[1-9]\d?)\)?)?[\-\.\ \\\/]?'  # TODO fix for md links
                                    + r'((?:\(?\d{1,}\)?[\-\.\ \\\/]?){0,}'
                                    + r'(?:#|ext\.?|extension|x)?[\-\.\ \\\/]?\d+)?)'
            }


def _get_differences(different_lines):
    differences = []
    for line in different_lines:
        match_diff_line = re.search(PATTERNS['different_line'], line)
        if line.startswith('+') and match_diff_line:
            # get only the lines that differ from the lines on the base page

            line = match_diff_line.group(1)
            for tag in re.findall(PATTERNS['unwanted_tag__has_placeholder'], line, flags=re.DOTALL):
                line = line.replace(tag, '')

            # this is not documented 'html2text' usage
            # 'html2text.html2text' does:
            # ```
            #   h = HTML2Text(baseurl=baseurl)
            #   return h.handle(html)
            # ```
            # and since 'body_width' can be only provided from CLI, default value (78) is used
            h2t = HTML2Text()
            h2t.body_width = 0
            line = h2t.handle(line)  # convert to md

            line = line.replace('*', '')
            line = line.strip()
            if line not in ['', '#']:
                lines = line.split('\n')
                lines = [line.strip() for line in lines]
                differences.extend(lines)
    return differences


def _extract_differences(base_page, other_page):
    # Compare the pages using difflib
    different_lines = difflib.unified_diff(
        str(base_page).splitlines(),
        str(other_page).splitlines(),
        lineterm='',
    )
    return _get_differences(different_lines)


def guess_name(parts, threshold=2, must_find=False):
    '''Guess person's name based on frequency of words

    Parameter threshold represents number of occurences of 2 words
    after which they will be considered as first and last name.
    '''
    parts_flat = [part for part in parts if isinstance(part, str)]
    for part in parts:
        if isinstance(part, dict):
            parts_flat.extend(*part.items())

    # preprocess data
    words = []
    for part in parts_flat:
        splitted = re.split(PATTERNS['repetetive_punct'], part)
        words += [word for word in splitted if word != '']

    # create a dictionary to count word pairs
    word_pair_original = dict()
    word_pair_counts = defaultdict(int)
    for i in range(len(words) - 1):
        if words[i] == '' or words[i+1] == '':
            continue
        word_pair = (words[i].lower(), words[i+1].lower())
        word_pair_original[word_pair] = (words[i], words[i+1])
        word_pair_counts[word_pair] += 1

    # sort the dictionary by counts in descending order
    sorted_word_pairs = sorted(
        word_pair_counts.items(),
        key=lambda x: x[1],
        reverse=True
    )

    # filter and return first and last name word pairs
    for word_pair, count in sorted_word_pairs:
        og_first_name = word_pair_original[word_pair][0]
        og_last_name = word_pair_original[word_pair][1]
        if og_first_name[0].isupper() and og_last_name[0].isupper():
            if count >= threshold:
                return ' '.join(word_pair).title()
    if not must_find or len(sorted_word_pairs) == 0:
        return None
    # choose most occuring pair as person's name
    full_name = sorted_word_pairs[0][0]
    return f'{full_name[0]} {full_name[1]}'.title()


def _update_context(context, extracted_info, replacement):
    # remove found info from context
    if isinstance(extracted_info, str):
        extracted_info = [extracted_info]
    for info in extracted_info:
        context = context.replace(info, replacement)

    return context


def extract_international_phone_numbers(text, country_code):
    number_info = {'numbers': [], 'context': text}
    phone_numbers = []
    for number in PhoneNumberMatcher(text, country_code):
        phone_numbers.append(number.raw_string)
    if len(phone_numbers) > 0:
        number_info['context'] = _update_context(text, phone_numbers, 'PHONE_NUMBER')
        for i, number in enumerate(phone_numbers):
            number = parse(number, country_code)
            try:
                phone_numbers[i] = format_number(number, PhoneNumberFormat.E164)
            except Exception:
                pass
        number_info['numbers'] = phone_numbers
    return number_info


def extract_national_phone_numbers(text):
    number_info = {'numbers': [], 'context': text}
    phone_numbers = []
    match_number = re.findall(PATTERNS['number'], text)
    if match_number:
        phone_numbers = [num_match[0] for num_match in match_number if num_match[0] != '']
    # add numbers if country code is not present
    if len(phone_numbers) > 0:
        number_info['numbers'] = phone_numbers
        number_info['context'] = _update_context(text, phone_numbers, 'PHONE_NUMBER')
    return number_info


def extract_phone_numbers(text, country_code):  # TODO fix for national nums, e.g. '011/1234-567'
    number_info = {'numbers': [], 'context': text}
    international_numbers = extract_international_phone_numbers(text, country_code)
    number_info['numbers'].extend(international_numbers['numbers'])
    number_info['context'] = international_numbers['context']
    return number_info


def __process_links(match_md_link, resume_links, resume_emails, context):
    found_something = True
    for md_link in match_md_link:
        link = md_link[1].strip()
        if 'mailto:' in link:
            if link in resume_emails:
                found_something = False
            else:
                context = _update_context(context, f'[{md_link[0]}]({md_link[1]})', 'EMAIL')
                link = link.replace('mailto:', '').strip()
                resume_emails.append(link)
        else:
            url_parts = urllib.parse.urlsplit(link)
            encoded_query = urllib.parse.quote(url_parts.query, safe='=&')
            link = urllib.parse.urlunsplit(url_parts._replace(query=encoded_query))
            # find key
            key = md_link[0].strip()
            if is_url(key):
                key = to_key(link)
            # check if it is relative link
            elif re.search(r'^[^(www)|(http)].+?\..+(/.+)*$', link.lower()):
                key = 'this'
                # check if it is relative link to an image
                if re.search(r'^(images?/)?.*?\.(jpg|jpeg|png|svg|webp|gif|bmp|ppm)$', link.lower()):
                    key = 'images'

            if key == '':
                key = 'profile_picture'

            # add new link
            if key not in resume_links:
                resume_links[key] = link
            elif isinstance(resume_links[key], str):
                resume_links[key] = [link, resume_links[key]]
            else:
                resume_links[key].append(link)
            context = _update_context(context, f'[{md_link[0]}]({md_link[1]})', 'LINK')
    return {
        'found_something': found_something,
        'context': context}


def _post_processing(resume, name_candidates):
    if 'Source URL' in resume:
        resume['url'] = resume.pop('Source URL')
    # try to guess person's name
    possible_name = guess_name(name_candidates, must_find=True)
    if possible_name is not None:
        name = possible_name
        if 'Source text' in resume:
            link_text = resume.pop('Source text')
            name_ci = longest_common_substring(link_text, possible_name, case_sensitive=False)
            name_parts = name_ci.split()
            name = ' '.join([name_part.capitalize() for name_part in name_parts])
        resume['name'] = name
        # remove name instances from `other`
        if possible_name in resume['other']:
            resume['other'].remove(possible_name)
        if possible_name.upper() in resume['other']:
            resume['other'].remove(possible_name.upper())
        # check if profile image link is missing
        if 'profile_picture' not in resume['links']:
            first, last = possible_name.lower().split()
            if 'images' in resume['links']:
                for img_link in resume['links']['images']:
                    if 'images' in resume and (
                        first.lower() in img_link.lower() or last.lower() in img_link.lower()
                    ):
                        resume['links']['profile_picture'] = img_link
                        if isinstance(resume['links']['images'], str):
                            del resume['links']['images']
                        else:
                            resume['links']['images'].remove(img_link)
                        break
        # convert relative links to abs
        if 'url' in resume:
            url = resume['url']
            for key in ['this', 'images', 'profile_picture']:
                if key in resume['links']:
                    if key == 'profile_picture':
                        print(resume['links'][key])
                    if isinstance(resume['links'][key], str):
                        resume['links'][key] = to_abs_path('/' + resume['links'][key], url)
                    else:
                        resume['links'][key] = [to_abs_path('/' + link, url) for link in resume['links'][key]]
        # remove duplicates
        resume['emails'] = list(dict.fromkeys(resume['emails']))
        resume['other'] = list(dict.fromkeys(resume['other']))
        return resume


def _parse_differences(differences, country_code=None):
    '''returns resume information which is extracted from differences between pages'''
    resume = {
        'context': [],
        'emails': [],
        'links': dict(),
        'other': [],
        'phone_numbers': []
        }
    name_candidates = []
    for difference in differences:
        match_email = re.findall(PATTERNS['email'], difference)
        match_md_link = re.findall(PATTERNS['md_link'], difference)
        match_label_field = re.search(PATTERNS['label_field'], difference)
        match_label_field_with_value = re.search(PATTERNS['label_field_with_value'], difference)

        found_something = False
        context = difference
        # add numbers to resume
        number_info = extract_phone_numbers(difference, country_code)
        if len(number_info['numbers']) > 0:
            resume['phone_numbers'] = number_info['numbers']
            context = number_info['context']
            found_something = True
        # add links to resume
        if match_md_link:
            result = __process_links(match_md_link, resume['links'], resume['emails'], context)
            context = result['context']
            found_something = result['found_something']
        # add emails to resume
        # note: this has to go after md link match to avoid matching the same thing multiple times
        if match_email:
            found_something = True
            for email in match_email:
                resume['emails'].append(email)
                name_candidates.append(email)
                context = _update_context(context, email, 'EMAIL')
        # add anything else to resume
        if not found_something:
            if match_label_field_with_value:
                # add key-value pair as top-level info
                key = match_label_field_with_value.group(1)
                value = match_label_field_with_value.group(2)
                resume[key.strip()] = value.strip()
                context = _update_context(context, key, 'FIELD_KEY')
                context = _update_context(context, value, 'FIELD_VAL')
            elif not match_label_field:
                # add the rest
                resume['other'].append(difference)
                name_candidates.append(difference)
        if context != difference:
            resume['context'] += [context]
    final_resume = _post_processing(resume, name_candidates)
    return final_resume


def get_resume_info(base_page, page, country_code=None):
    '''Extract resume information by comparing to base page

    This is based on the fact that the pages on the same website mostly have the same layout.
    Only the fields for information were modified when comparing two profiles.

    Returns dictionary with identified and not identified field values
    '''
    differences = _extract_differences(base_page, page)
    return _parse_differences(differences, country_code)


def get_resumes(pages, country_code=None):
    assert isinstance(pages, list), 'Parameter \'pages\' must be an list instance'
    assert len(pages) > 1, 'Number of pages has to be greater then 1'

    parsed_pages = []
    for page in pages:
        soup = BeautifulSoup(page, 'html.parser')
        # remove javascipt and css code
        for tag in soup.find_all("script"):
            soup.script.decompose()
        for tag in soup.find_all("style"):
            soup.style.decompose()
        parsed_pages.append(soup)
    base_page = parsed_pages[0]
    other_pages = parsed_pages[1:]
    resumes = dict()

    # if country_code is not set, try to infer it from html
    if country_code is None:
        country_code = base_page.html.get('lang')
        country_code = country_code.upper() if country_code is not None else None

    # charset = base_page.meta.get('charset')

    for i, page in enumerate(other_pages):
        key = str(i)
        resumes[key] = get_resume_info(base_page, page, country_code)

    key = len(resumes)
    resumes[key] = get_resume_info(other_pages[0], base_page, country_code)
    return resumes


def export_resumes(resumes, export_path, export_method):
    if export_method == 'json':
        with open(os.path.join(export_path, 'resumes.json'), 'w') as f:
            f.write(resumes)


def get_resumes_from_dir(dir_path, export_path=None, export_method='json'):
    profile_pages = []
    for root, _, files in os.walk(dir_path):
        for file in files:
            with open(os.path.join(root, file), 'r') as f:
                profile_pages.append(f.read())
    resumes = get_resumes(profile_pages)
    pretty_resumes = json.dumps(resumes, indent=4)
    if export_path is not None:
        export_resumes(''.join(pretty_resumes), export_path, export_method)
    return pretty_resumes


def longest_common_substring(str1, str2, case_sensitive=True):
    m = len(str1)
    n = len(str2)
    if not case_sensitive:
        str1 = str1.lower()
        str2 = str2.lower()
    # create a table to store the lengths of common substrings
    table = [[0] * (n + 1) for _ in range(m + 1)]
    # variables to keep track of the longest common substring
    max_length = 0
    end_index = 0
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if str1[i - 1] == str2[j - 1]:
                table[i][j] = table[i - 1][j - 1] + 1
                if table[i][j] > max_length:
                    max_length = table[i][j]
                    end_index = i
    # extract the longest common substring
    longest_substring = str1[end_index - max_length: end_index]
    return longest_substring



import re
import os
import sys
import random
import tldextract
from urllib.parse import urlparse, quote_plus

from dataclasses import dataclass

from profilescout.common.constants import ConstantsNamespace
from profilescout.common.exceptions import LongFilenameException


constants = ConstantsNamespace


@dataclass
class PageLink:
    url: str
    depth: int
    parent_url: str = None
    txt: str = ''


def is_url(s):
    extracted = tldextract.extract(s)
    return bool(extracted.domain and extracted.suffix)


def is_valid(url, base_url):
    if not isinstance(url, str):
        return False

    if url == '.' or url == '' or url[-1] == '#':
        return False

    if url.startswith('mailto:') or url.startswith('tel:') or url.startswith('javascript:'):
        return False

    if '?nocache' in url:
        return False

    # search for url with filename and extension and check if exntension is invalid
    matched = re.search(r"^(https?://)?(www\.)?[a-zA-Z0-9_-]+(\.[a-zA-Z]{2,6})+(/\S*)*/\S+\.(\w{1,5})$", url)
    if matched:
        ext = matched.groups()[-1]
        if ext.lower() in constants.INVALID_EXTENSIONS:
            return False

    base_extract = tldextract.extract(base_url)
    link_extract = tldextract.extract(url)
    base_subdo_reversed = base_extract.subdomain.split('.')[::-1]
    link_subdo_reversed = link_extract.subdomain.split('.')[::-1]
    common_len = min(len(base_subdo_reversed), len(link_subdo_reversed))
    # TODO rethink; could cause problem with some sites that have profiles
    # on another url or as PDF/DOCX/...
    has_protocol = url.startswith('http://') or url.startswith('https://')
    if (
        has_protocol
        and (  # check if they have common subdomain (base subdomain must be contained in url's subdomain)
            base_extract.domain != link_extract.domain
            or any(base_subdo_reversed[i] != link_subdo_reversed[i] for i in range(common_len))
        )
    ):
        return False

    return True


def to_abs_path(url, current_url):
    base_url = to_base_url(current_url)
    # fix relative links or links that start with '/'
    abs_url = url
    if url.startswith('http'):
        abs_url = url
    elif url.startswith('www'):
        abs_url = 'http://' + url
    else:
        # relative path or absoulte path from '/'
        if url[0] == '/':
            # absolute path
            if base_url[-1] == '/':
                url = url[1:]  # remove '/' since base_url ends with one
            abs_url = base_url + url
        else:
            # relative path
            if current_url[-1] != '/':
                url = '/' + url
            abs_url = current_url + url
    return abs_url


def to_key(url):
    result = tldextract.extract(url)
    if result.subdomain in ['', 'www']:
        return result.domain
    if 'www' in result.subdomain:
        subdomain = result.subdomain.replace('www.', '')
    return f'{result.domain}-{subdomain}'


def to_fqdn(url):
    return tldextract.extract(url).fqdn


def to_base_url(url):
    # extract base url from given url
    base_url = to_fqdn(url)
    if 'https://' in url:
        base_url = 'https://' + base_url
    else:
        base_url = 'http://' + base_url
    return base_url + '/'


def remove_duplicates(page_links):
    unique_links = page_links

    i = 0
    n = len(unique_links)
    while i < n-1:
        j = i+1
        while j < n:
            if unique_links[i].url == unique_links[j].url:
                # update depth if there is better path
                if unique_links[i].depth > unique_links[j].depth:
                    unique_links[i].depth = unique_links[j].depth

                # delete duplicate
                del unique_links[j]
                j -= 1
                n -= 1
            j += 1
        i += 1

    return unique_links


def prioritize_relevant(link_queue):
    '''move relevant word at the beginning of the queue'''
    front = []
    rest = []
    for page_link in link_queue:
        has_relevant = False
        for word in constants.RELEVANT_WORDS:
            word = quote_plus(word)
            if word in page_link.url.lower():
                front.append(page_link)
                has_relevant = True
                break
        if not has_relevant:
            rest.append(page_link)

    return front + rest


def with_and_without_www(url):
    parsed_url = urlparse(url)
    if parsed_url.scheme == '' or parsed_url.hostname is None:
        if url.startswith('www.'):
            return url, url.replace('www.', '')
        return f'www.{url}', url
    if parsed_url.hostname.startswith('www'):
        return url, url.replace('www.', '')
    return url.replace(f'{parsed_url.scheme}://', f'{parsed_url.scheme}://www.'), url


def filter_out_invalid(page_links, base_url):
    result = filter(lambda pl: is_valid(pl.url, base_url),
                    page_links)
    return list(result)


def filter_out_visited(page_links, visited_links):
    result = []
    for page_link in page_links:
        www, wo_www = with_and_without_www(page_link.url)
        if www not in visited_links and wo_www not in visited_links:
            result += [page_link]
    return result


def filter_out_present_links(page_links, links_to_visit):
    result = []
    urls_to_visit = [to_visit.url for to_visit in links_to_visit]
    for page_link in page_links:
        www, wo_www = with_and_without_www(page_link.url)
        if www not in urls_to_visit and wo_www not in urls_to_visit:
            result += [page_link]
    return result


def filter_out_long(page_links, err_file=sys.stderr):
    # exclude links that are overly long
    not_too_long = []
    for page_link in page_links:
        if len(page_link.url) > 310:
            print(f'WARN: url={page_link.url!r} is too long and may not be relevant. Ignored',
                  file=err_file)
        else:
            not_too_long.append(page_link)

    return not_too_long


def to_filename(url, export_path, extension):
    filename = url

    # remove '/' at the end
    if filename[-1] == '/':
        filename = filename[:-1]

    filename = filename.replace(f'__.{extension}', f'.{extension}')
    filename = filename.replace('http://', '').replace('https://', '')

    for unsafe_part in constants.CHAR_REPLACEMENTS.keys():
        if unsafe_part in filename:
            filename = filename.replace(unsafe_part, constants.CHAR_REPLACEMENTS[unsafe_part])

    # check if filename exceeds upper limit for number of characters
    limit = os.pathconf(export_path, 'PC_NAME_MAX')
    limit = min(limit, constants.FILENAME_MAX_LENGHT)
    if len(filename) + len(extension) + 1 > limit:
        raise LongFilenameException(filename, limit)
    else:
        filename += '.' + extension

    return filename


def url2file_path(link, export_path, extension, err_file=sys.stderr):
    filename = ''
    try:
        filename = to_filename(link, export_path, extension)
    except LongFilenameException as lfe:
        filename = lfe.args[0]

        # cut off chars that exceed limit
        filename = filename[:lfe.limit]

        # replace the remaining string with suffix and extension
        ext = '.' + extension
        suffix = constants.FILENAME_CUT_SUFFIX
        suffix += str(random.randint(1000, 9999))
        overwrite_len = len(suffix) + len(ext)
        filename = filename[:-overwrite_len] + suffix + ext
        print('WARN: Link was too long.',
              f'The filename of has changed to: {filename}',
              file=err_file)
    path = os.path.join(export_path, filename)
    if os.path.exists(path):
        print(f'WARN: File already exists at: {path}', file=err_file)
        return None

    return path


def replace_param_vals(url, replacement='####'):
    return re.sub(r'(?<=[?&])(.*?)=(.*?)(?=&|$)', r'\1='+replacement, url)


def _create_format_freq_tuple(fmt_freq, placeholder='####'):
    fmt_path_part = fmt_freq[0]
    qmark_idx = fmt_path_part.find('?')
    if qmark_idx != -1:
        fmt_path_part = fmt_path_part[:qmark_idx]
    query_part = fmt_freq[0][qmark_idx:]
    query_var_count = query_part.count(placeholder)
    # penalize those formats which don't have any variable part in url query
    query_var_count_asc = query_var_count if query_var_count > 0 else sys.maxsize
    fmt_freq_desc = -fmt_freq[1]
    var_count_asc = fmt_freq[0].count(placeholder) if fmt_freq[0].count(placeholder) > 0 else sys.maxsize
    has_slash_desc = '/' not in fmt_freq[0]
    has_query_desc = '?' not in fmt_freq[0]
    has_var_in_path_desc = placeholder not in fmt_path_part
    return has_slash_desc, has_query_desc, has_var_in_path_desc, var_count_asc, query_var_count_asc, fmt_freq_desc


def most_common_format(urls, placeholder='####'):
    parsed_urls = [urlparse(url) for url in urls]
    urls_without_domain = [f'{url.path}?{url.query}'
                           if url.query != '' else url.path
                           for url in parsed_urls]
    encoded_urls = [replace_param_vals(url, placeholder) for url in urls_without_domain]
    fmts = dict()
    for url1 in encoded_urls:
        url1_parts = url1.split('/')
        url1_parts = [part for part in url1_parts if part != '']
        for url2 in encoded_urls:
            url2_parts = url2.split('/')
            url2_parts = [part for part in url2_parts if part != '']
            fmt_parts = []
            lmin = min(len(url1_parts), len(url2_parts))
            for i in range(lmin):
                fmt_parts += [url1_parts[i]] if url1_parts[i] == url2_parts[i] else [placeholder]
            lmax = max(len(url1_parts), len(url2_parts))
            fmt_parts += [placeholder] * (lmax - lmin)
            # if placeholder in fmt_parts or len(fmt_parts) == 1:
            fmt = '/'.join(fmt_parts)
            fmts[fmt] = fmts.get(fmt, 0) + 1
    if len(fmts) == 0:
        return None
    # sort formats by count of '#' in format, negative value of '?' freqency
    # and negative value of placeholder frequency
    # note: negative value is used in order to achive desc ordering by frequency
    fmt_freqs = sorted(fmts.items(), key=lambda x: _create_format_freq_tuple(x, placeholder))
    common_fmt = fmt_freqs[0][0]
    parsed_url = urlparse(urls[0])
    base_url = f"{parsed_url.scheme}://{parsed_url.netloc}/"
    return base_url + common_fmt


def match_profile_fmt(url, fmt, placeholder):
    assert placeholder in fmt, f'expected a format containing the placeholder {placeholder!r}, but recieved: {fmt!r}'
    fmt_escaped = re.escape(fmt)
    fmt_pattern = fmt_escaped.replace(re.escape(placeholder), r'.+?')
    return re.search(fmt_pattern, url)


def is_valid_sublink(url, fmt, placeholder):
    return fmt is None or match_profile_fmt(url, fmt, placeholder)



from abc import ABC, abstractmethod


class DetectionStrategy(ABC):
    """Interface/Base class for profile detection strategy"""

    @abstractmethod
    def successful(self):
        pass

    @abstractmethod
    def get_result(self):
        pass
        pass

    @abstractmethod
    def reset(self):
        pass



from dataclasses import dataclass, field

from profilescout.common.constants import ConstantsNamespace
from profilescout.common.interfaces import DetectionStrategy
from profilescout.link.utils import most_common_format


constants = ConstantsNamespace


@dataclass
class OriginPageDetectionStrategy(DetectionStrategy):
    succeeded: bool = False
    result: dict = None
    origin_candidates: dict = field(default_factory=dict)

    def successful(self):
        return self.succeeded

    def get_result(self):
        return self.result

    def reset(self):
        self.succeeded = False
        self.result = None
        self.origin_candidates = dict()  # TODO determine whether or not this should be completely cleared

    def analyse(self, curr_page, classifier_name, resolution):
        action_result = curr_page.is_profile(classifier_name, *resolution)
        profile_detected = action_result.val
        if action_result.successful and profile_detected:
            # assume that this is initial page
            origin = curr_page.link.url
            parent_url = curr_page.link.parent_url
            if parent_url is not None:
                origin = parent_url
            if origin not in self.origin_candidates:
                self.origin_candidates[origin] = [curr_page.link.url]
            else:
                self.origin_candidates[origin] += [curr_page.link.url]

            # check if the profile page origin is found
            children = self.origin_candidates[origin]
            children_count = len(children)
            if children_count == constants.ORIGIN_PAGE_THRESHOLD:
                self.result = {
                    'origin': origin,
                    'depth': curr_page.link.depth - 1,
                    'most_common_format': most_common_format(children),  # TODO add placeholder
                    'message': f'Found profile page origin at {origin!r}'}
                self.succeeded = True
        return self.result



import os
import sys
import textwrap
from concurrent.futures import ThreadPoolExecutor, wait

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from profilescout.__about__ import __version__
from profilescout.common.constants import ConstantsNamespace
from profilescout.link.utils import to_fqdn, to_base_url
from profilescout.web.webpage import WebpageActionType, ScrapeOption
from profilescout.web.crawl import CrawlOptions, crawl_website
from profilescout.classification.classifier import CLASSIFIERS_DIR
from profilescout.extraction.htmlextract import get_resumes_from_dir


constants = ConstantsNamespace


def generate_crawl_inputs(
        url, urls_file_path, export_path,
        crawl_sleep, depth, max_pages, max_threads,
        include_fragment, bump_relevant, peserve_uri, use_buffer,
        action_type, scrape_option,
        resolution, image_classifier):
    user_inputs = []
    crawl_inputs = []
    if urls_file_path is None:
        user_inputs += [(url, depth, crawl_sleep)]
    else:
        # read urls and other data from file
        lines = []
        with open(urls_file_path, 'r') as f:
            lines = f.readlines()
        for line in lines:
            # set default or passed values from command line in case
            # they are not present in the file for given line
            read_depth, read_crawl_sleep = depth, crawl_sleep
            # extract fields
            line = line.replace('\n', '').strip()
            parts = line.split(' ')
            if len(parts) == 1 and parts[0] != '':
                read_url = parts[0]
            elif len(parts) == 2:
                read_depth, read_url = parts[0], parts[1]
            elif len(parts) == 3:
                read_depth, read_crawl_sleep, read_url = parts[0], parts[1], parts[2]
            else:
                print(f'WARN: {line=} is not in valid format. Ignored')
                continue
            if not read_url.startswith('http://') and not read_url.startswith('https://'):
                print(f"WARN: {read_url=} should start with 'http://' or 'https://'. Ignored")
                continue  # TODO add option to choose action in this situation
            user_inputs += [(read_url, read_depth, read_crawl_sleep)]
    # if there are too many links, buffer output to avoid load on storage device
    if len(user_inputs) > constants.BUFF_THRESHOLD:
        use_buffer = True
    # prepare crawl inputs
    for read_url, read_depth, read_crawl_sleep in user_inputs:
        scraping = True
        output_fname = to_fqdn(read_url)
        export_path_for_url = os.path.join(export_path, output_fname)
        if action_type == WebpageActionType.FIND_ORIGIN:
            scraping = False
        # craft crawl input
        crawl_options = CrawlOptions(
            max_depth=read_depth,
            max_pages=max_pages,
            crawl_sleep=read_crawl_sleep,
            include_fragment=include_fragment,
            bump_relevant=bump_relevant,
            use_buffer=use_buffer,
            scraping=scraping,
            resolution=resolution,
            classifier_name=image_classifier)
        if not peserve_uri:
            read_url = to_base_url(read_url)
        crawl_inputs += [(
            export_path_for_url,
            read_url,
            crawl_options,
            action_type,
            scrape_option)]
    return crawl_inputs


def main(url, urls_file_path, export_path, directory,
         crawl_sleep, depth, max_pages, max_threads,
         include_fragment, bump_relevant, peserve_uri, use_buffer,
         action_type, scrape_option,
         resolution, image_classifier):
    # check if info extraction is chosen
    if directory is not None:
        if export_path == '':
            export_path = None
        resumes = get_resumes_from_dir(directory, export_path)
        print(resumes)
        return
    crawl_inputs = generate_crawl_inputs(
        url, urls_file_path, export_path,
        crawl_sleep, depth, max_pages, max_threads,
        include_fragment, bump_relevant, peserve_uri, use_buffer,
        action_type, scrape_option,
        resolution, image_classifier)
    # crawl each website in seperate thread
    print(f'INFO: PID: {os.getpid()!r}')
    print('INFO: Start submitting URls for crawling...')
    with ThreadPoolExecutor(max_workers=max_threads) as executor:
        # Submit each URL for crawling
        futures = [executor.submit(crawl_website, *crawl_input) for crawl_input in crawl_inputs]
        print('INFO: Waiting threads to complete...')
        # Wait for all tasks to complete
        wait(futures)
        print('INFO: Threads have completed the crawling')


def cli():
    import argparse

    scrape_choices = [so.name.lower() for so in list(ScrapeOption)]
    default_scrape_choice = ScrapeOption.ALL.name.lower()

    action_choices = [at.name.lower() for at in list(WebpageActionType)]
    action_choices.remove(WebpageActionType.UNKNOWN.name.lower())
    default_action_choice = WebpageActionType.SCRAPE_PAGES.name.lower()

    parser = argparse.ArgumentParser(
        description='Crawl website and do something with for each page',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent(f'''\
            Additional details
            ------------------
            Full input line format is: '[DEPTH [CRAWL_SLEEP]] URL"

            DEPTH and CRAWL_SLEEP are optional and if a number is present it will be consider as DEPTH.
            For example, "3 https://example.com" means that the URL should be crawled to a depth of 3.

            If some of the fields (DEPTH or CRAWL_SLEEP) are present in the line then corresponding argument is ignored.

            Writing too much on the storage drive can reduce its lifespan. To mitigate this issue, if there are more than
            {constants.BUFF_THRESHOLD} links, informational and error messages will be buffered and written at the end of
            the crawling process.

            RELEVANT_WORDS={constants.RELEVANT_WORDS}
            '''))
    input_group = parser.add_mutually_exclusive_group(required=True)
    input_group.add_argument(
        '--url',
        help='URL of the website to crawl',
        dest='url')
    input_group.add_argument(
        '-f', '--file',
        help='Path to the file with URLs of the websites to crawl',
        dest='urls_file_path')
    input_group.add_argument(
        '-D', '--directory',
        help="Extract data from HTML files in the directory. To avoid saving output, set '-ep'/'--export-path' to ''",
        dest='directory')

    parser.add_argument(
        '-a', '--action',
        help="Action to perform at a time of visiting the page (default: %(default)s)",
        dest='action',
        choices=action_choices, default=default_action_choice,)
    parser.add_argument(
        '-b', '--buffer',
        help="Buffer errors and outputs until crawling of website is finished and then create logs",
        dest='use_buffer',
        action='store_const', const=True, default=False)
    parser.add_argument(
        '-br', '--bump-relevant',
        help="Bump relevant links to the top of the visiting queue (based on RELEVANT_WORDS list)",
        dest='bump_relevant',
        action='store_const', const=True, default=False)
    parser.add_argument(
        '-ep', '--export-path',
        help='Path to destination directory for exporting',
        dest='export_path',
        default='./results')
    parser.add_argument(
        '-ic', '--image-classifier',
        help="Image classifier to be used for identifying profile pages (default: %(default)s)",
        dest='image_classifier',
        choices=constants.IMAGE_CLASSIFIERS, default=constants.IMAGE_CLASSIFIERS[0])
    parser.add_argument(
        '-cs', '--crawl-sleep',
        help='Time to sleep between each page visit (default: %(default)s)',
        dest='crawl_sleep',
        default=2, type=int)
    parser.add_argument(
        '-d', '--depth',
        help='Maximum crawl depth (default: %(default)s)',
        dest='depth',
        default=2, type=int)
    parser.add_argument(
        '-if', '--include-fragment',
        help="Consider links with URI Fragment (e.g. http://example.com/some#fragment) as seperate page",
        dest='include_fragment',
        action='store_const', const=True, default=False)
    parser.add_argument(
        '-ol', '--output-log-path',
        help="Path to output log file. Ignored if '-f'/'--file' is used",
        dest='out_log_path')
    parser.add_argument(
        '-el', '--error-log-path',
        help="Path to error log file. Ignored if '-f'/'--file' is used",
        dest='err_log_path')
    parser.add_argument(
        '-so', '--scrape-option',
        help="Data to be scraped (default: %(default)s)",
        dest='scrape_option',
        choices=scrape_choices, default=default_scrape_choice)
    parser.add_argument(
        '-t', '--threads',
        help="Maximum number of threads to use if '-f'/'--file' is provided (default: %(default)s)",
        dest='max_threads',
        default=4, type=int)
    parser.add_argument(
        '-mp', '--max-pages',
        help='''
                Maximum number of pages to scrape
                and page is considered scraped if the action is performed successfully (default: unlimited)
                ''',
        dest='max_pages',
        type=int)
    parser.add_argument(
        '-p', '--preserve',
        help="Preserve whole URI (e.g. \'http://example.com/something/\' instead of  \'http://example.com/\')",
        dest='peserve_uri',
        action='store_const', const=True, default=False)
    parser.add_argument(
        '-r', '--resolution',
        help="Resolution of headless browser and output images. Format: WIDTHxHIGHT (default: %(default)s)",
        dest='resolution',
        default=f'{constants.WIDTH}x{constants.HEIGHT}')

    args = parser.parse_args()

    if args.url is not None:
        if not (args.url.startswith('http://')
                or args.url.startswith('https://')):
            parser.error("url should start with 'http://' or 'https://'")
            sys.exit()

    # validate resolution
    args.resolution = args.resolution.split('x')
    if (
        len(args.resolution) < 2
        or not args.resolution[0].isdigit()
        or args.resolution[0][0] == '0'
        or not args.resolution[1].isdigit()
        or args.resolution[1][0] == '0'
    ):
        parser.error("valid format for resolution: WIDTHxHEIGHT. Example: 2880x1620")
        sys.exit()

    # map action str to enum
    action_type = getattr(WebpageActionType, args.action.upper(), None)

    # map scrape option str to enum
    scrape_option = getattr(ScrapeOption, args.scrape_option.upper(), None)

    # create export dir if not present
    try:
        os.mkdir(args.export_path)
    except FileExistsError:
        print(f'INFO: Directory {args.export_path!r} already exists')
    except OSError as e:
        args.export_path = os.getcwd()
        parser.error(f'''
            ERROR: Unable create directory at {args.export_path!r}
            (reason: {e.strerror if hasattr(e, "strerror") else "unknown"}).
            The current directory has been set as the export directory
            ''')
        export_path = '.'

    if args.directory is None:
        # check for classifier dir and file presence
        try:
            if not os.listdir(CLASSIFIERS_DIR):
                h5_files = [file for file in os.listdir(CLASSIFIERS_DIR) if file.endswith('.h5')]

                if len(h5_files) == 0:
                    parser.error(f'Directory {CLASSIFIERS_DIR!r} does not contain any .h5 files')
                    sys.exit()

                if args.image_classifier not in h5_files:
                    parser.error(f'Model \'{args.image_classifier}.h5\' is not found at {CLASSIFIERS_DIR!r}')
                    sys.exit()
        except FileNotFoundError:
            parser.error(f'Directory {CLASSIFIERS_DIR!r} is not present')
            sys.exit()

    try:
        main(
            url=args.url,
            urls_file_path=args.urls_file_path,
            directory=args.directory,
            export_path=args.export_path,
            crawl_sleep=args.crawl_sleep,
            depth=args.depth,
            max_pages=args.max_pages,
            max_threads=args.max_threads,
            include_fragment=args.include_fragment,
            bump_relevant=args.bump_relevant,
            peserve_uri=args.peserve_uri,
            use_buffer=args.use_buffer,
            action_type=action_type,
            scrape_option=scrape_option,
            resolution=args.resolution,
            image_classifier=args.image_classifier)
    except KeyboardInterrupt:
        print('\nINFO: Exited')
    else:
        print('INFO: Finished')



import os
import time

from enum import Enum
from PIL import Image
from io import BytesIO

from profilescout.common.exceptions import WebDriverException, StaleElementReferenceException
from profilescout.common.constants import ConstantsNamespace
from profilescout.link.utils import PageLink, is_valid, url2file_path
from profilescout.classification.classifier import ImageClassifier


constants = ConstantsNamespace

WebpageActionType = Enum(
     'WebpageActionType', [
         'UNKNOWN',
         'SCRAPE_PAGES',
         'SCRAPE_PROFILES',
         'FIND_ORIGIN']
)

ScrapeOption = Enum('ScrapeOption', ['ALL', 'HTML', 'SCREENSHOT'])


class ActionResult:
    def __init__(self, successful, val=None, msg=''):
        self.successful = successful
        self.val = val
        self.msg = msg


class Webpage:
    def __init__(self, web_driver, page_link, out_file, err_file):
        self.__web_driver = web_driver
        self.link = page_link
        self.__out_file = out_file
        self.__err_file = err_file

    def visit(self):
        # navigate to the web page you want to capture
        try:
            self.__web_driver.get(self.link.url)
        except WebDriverException:
            # if not successful, retry after RETRY_TIME seconds
            time.sleep(constants.RETRY_TIME)

            try:
                self.__web_driver.get(self.link.url)
            except WebDriverException as e:
                raise e  # TODO

        # pause videos that have autoplay set to true
        pause_ap_script = 'videos = document.querySelectorAll("video"); for(video of videos) {video.pause()}'
        self.__web_driver.execute_script(pause_ap_script)

        content_type = self.__web_driver.execute_script("return document.contentType")

        if content_type.startswith('text'):
            return True

        return False

    def get_html(self):
        html = self.__web_driver.get_page_source()
        source_url_tag = f'<profilescout>Source URL:{self.link.url}</profilescout>'
        source_txt_tag = f'<profilescout>Source text:{self.link.txt}</profilescout>\n'
        return f'{source_url_tag}\n{source_txt_tag}\n{html}'

    def take_screenshot(self, width=constants.WIDTH, height=constants.HEIGHT):
        '''takes screenshot of current page and returns image as byte array'''
        self.__web_driver.set_window_size(width, height)

        # take a screenshot of the entire web page and store it in buffer
        screenshot_bytes = BytesIO(self.__web_driver.get_screenshot_as_png())  # TODO close
        image = Image.open(screenshot_bytes).convert("RGB")

        return ActionResult(True, image, 'Image is stored in a buffer')

    def scrape_page(self, export_path, scrape_option, width=constants.WIDTH, height=constants.HEIGHT):
        successful = True
        result = {'html': None, 'screenshot': None}
        self.__web_driver.set_window_size(width, height)

        if scrape_option in [ScrapeOption.ALL, ScrapeOption.SCREENSHOT]:
            # take a screenshot of the entire web page and save it as an image file
            path = url2file_path(
                self.link.url,
                os.path.join(export_path, 'screenshots'),
                constants.IMG_EXT,
                self.__err_file)
            if path is None:
                return ActionResult(False, 'Failed to craft valid storing path for the screenshot')
            result['screenshot'] = path
            successful = self.__web_driver.save_screenshot(path)

        if scrape_option in [ScrapeOption.ALL, ScrapeOption.HTML]:
            # save html as a file
            path = url2file_path(
                self.link.url,
                os.path.join(export_path, 'html'),
                'html',
                self.__err_file)
            if path is None:
                return ActionResult(False, 'Failed to craft valid storing path for the html')
            result['html'] = path
            html = self.get_html()
            with open(path, 'w') as f:
                f.write(html)

        return ActionResult(successful, result)

    def is_profile(self, classifier_name, width, height):
        result = self.take_screenshot(width, height)

        if not result.successful:
            return ActionResult(False, 'Inference was not successfully performed')

        img_bytes = result.val
        img_classifier = ImageClassifier(classifier_name)
        profile_detected = img_classifier.predict(img_bytes, height, width)

        if profile_detected:
            print(f'INFO: Detected as profile page: {self.link.url}', file=self.__out_file)

        return ActionResult(True, profile_detected, 'Inference was successfully performed')

    def extract_links(self, base_url, include_fragment=False, from_structure=False, previous_links=[]):
        page_links = []
        urls = []
        xpath = '//a[@href]'
        if from_structure:
            xpath = '''//*[self::table or self::ol or self::ul or self::section]//a[@href
                            and not(ancestor::header or ancestor::footer or ancestor::nav)
                            and not(ancestor::div[contains(@id, 'footer')])
                            and not(ancestor::div[contains(@id, 'nav')])
                            and not(ancestor::div[contains(@id, 'navigation')])
                            and not(ancestor::div[contains(@class, 'footer')])
                            and not(ancestor::div[contains(@class, 'nav')])
                            and not(ancestor::div[contains(@class, 'navigation')])
                            and not(ancestor::div[contains(@role, 'footer')])
                            and not(ancestor::div[contains(@role, 'nav')])
                            and not(ancestor::div[contains(@role, 'navigation')])
                        ]'''
        a_tags = self.__web_driver.find_elements_with_xpath(xpath)
        for a_tag in a_tags:
            href = ''
            try:
                href = a_tag.get_attribute('href').strip()
                txt = a_tag.text
            except StaleElementReferenceException:
                print(f'ERROR: One of the links to visit next, "{self.link.url}",',
                      'is skipped (reason: stale element)',
                      file=self.__err_file)
            except WebDriverException as e:
                print(f'ERROR: One of the links to visit next, "{self.link.url}",',
                      f'is skipped (reason: {str(e)})',
                      file=self.__err_file)
            else:
                idx = href.find('?nocache')
                if idx != -1:
                    href = href[:idx]

                if not include_fragment:
                    # remove fragment
                    frag_idx = href.find('#')
                    if frag_idx != -1:
                        href = href[:frag_idx]

                if href not in urls and is_valid(href, base_url):
                    urls.append(href)
                    page_link = PageLink(href, self.link.depth+1, self.link.url, txt)
                    if from_structure and page_link.url in previous_links:
                        continue
                    page_links += [page_link]

        return page_links



def parse_web_driver_exception(e, url):
    reason = 'error'
    err_msg = ''

    if 'ERR_NAME_NOT_RESOLVED' in str(e):
        reason = 'unresolved'
        err_msg = f'ERROR: cannot visit url "{url}" (reason: name cannot be resolved)'
    elif 'ERR_ADDRESS_UNREACHABLE' in str(e):
        reason = 'unreachable'
        err_msg = f'ERROR: cannot visit url "{url}" (reason: address cannot be reached)'
    elif 'ERR_CONNECTION_TIMED_OUT' in str(e):
        reason = 'timed out'
        err_msg = f'ERROR: cannot visit url "{url}" (reason: connection timed out)'
    elif 'stale element reference' in str(e):
        reason = 'stale'
        err_msg = f'ERROR: cannot visit url "{url}" (reason: stale element)'
    elif 'ERR_SSL_VERSION_OR_CIPHER_MISMATCH' in str(e):
        reason = 'https not supported'
        err_msg = f'ERROR: cannot visit url "{url}" (reason: site does not support https)'
    else:
        err_msg = f'ERROR: {str(e)}'

    return err_msg, reason


class LongFilenameException(Exception):
    def __init__(self, message, limit):
        super().__init__(message)
        self.limit = limit


class WebDriverException(Exception):
    """Custom exception to abstract WebDriverException from users."""

    def __init__(self, message, screen=None, stacktrace=None, filename=None, lineno=None):
        self.message = message
        self.screen = screen
        self.stacktrace = stacktrace
        self.filename = filename
        self.lineno = lineno
        super().__init__(self.message)

    @staticmethod
    def from_webdriver_exception(e):
        return WebDriverException(
            message=str(e),
            screen=getattr(e, 'screen', None),
            stacktrace=getattr(e, 'stacktrace', None),
            filename=getattr(e, 'filename', None),
            lineno=getattr(e, 'lineno', None)
        )


class StaleElementReferenceException(Exception):
    """Custom exception to abstract StaleElementReferenceException from users."""

    def __init__(self, message, screen=None, stacktrace=None, filename=None, lineno=None):
        self.message = message
        self.screen = screen
        self.stacktrace = stacktrace
        self.filename = filename
        self.lineno = lineno
        super().__init__(self.message)

    @staticmethod
    def from_stale_element_exception(e):
        return StaleElementReferenceException(
            message=str(e),
            screen=getattr(e, 'screen', None),
            stacktrace=getattr(e, 'stacktrace', None),
            filename=getattr(e, 'filename', None),
            lineno=getattr(e, 'lineno', None)
        )



import os

from profilescout.common.constants import ConstantsNamespace

import numpy as np
import tensorflow as tf
from tensorflow.keras.models import load_model


constants = ConstantsNamespace

CLASSIFIERS_DIR = os.path.abspath(
    os.path.join(__file__,
                 os.pardir,
                 os.pardir,
                 os.pardir,
                 'classifiers'))


class ImageClassifier:

    __model = None

    def __init__(self, classifier_name):
        self.__classifier_name = classifier_name

        path = os.path.join(CLASSIFIERS_DIR,  f'{classifier_name}.h5')
        self.__classifier_path = path

    def __load_model(self):
        if self.__model is None:
            self.__model = load_model(self.__classifier_path)
        return self.__model

    def __preprocess_scooby(self, image, resize_height=480, resize_width=360):
        image = tf.image.rgb_to_grayscale(image)
        resized = tf.image.resize(image, [resize_height, resize_width])
        reshaped = tf.reshape(resized, (1, resize_height, resize_width, 1))
        normalized = tf.divide(reshaped, 255.0)

        return normalized

    def __preprocess_batman(self, image, resize_height=360, resize_width=480):
        image = tf.image.rgb_to_grayscale(image)
        resized = tf.image.resize(image, [resize_height, resize_width])
        reshaped = tf.reshape(resized, (1, resize_height, resize_width, 1))
        normalized = tf.divide(reshaped, 255.0)

        return normalized

    def predict(self, image, height, width, channels=3, verbose=0):
        self.__load_model()

        array = np.array(image)

        # Ensure the array has 3 channels
        array = array.reshape((1,) + array.shape)

        tensor = tf.convert_to_tensor(array, dtype=tf.uint8)

        # TODO refactor
        preprocessed_image = tensor
        if 'scooby' in self.__classifier_name:
            preprocessed_image = self.__preprocess_scooby(tensor)
        elif 'batman' in self.__classifier_name:
            preprocessed_image = self.__preprocess_batman(tensor)

        is_profile_percetage = self.__model(preprocessed_image, training=False)
        if is_profile_percetage > constants.PREDICTION_THRESHOLD:
            return True

        return False



import os
import sys
import time
import random
import traceback
import copy

from io import StringIO
from dataclasses import dataclass
from http.client import RemoteDisconnected

from profilescout.common.constants import ConstantsNamespace
from profilescout.common.structures import OriginPageDetectionStrategy
from profilescout.link.utils import is_valid_sublink
from profilescout.web.manager import CrawlManager, CrawlStatus
from profilescout.web.webdriver import setup_web_driver
from profilescout.web.webpage import ScrapeOption, WebpageActionType


constants = ConstantsNamespace


def _close_everything(web_driver, out_file, err_file, export_path, use_buffer):
    if web_driver is not None:
        web_driver.quit()
    # make sure that everything is flushed before stream is closed
    out_file.flush()
    err_file.flush()
    if use_buffer:
        out_content = out_file.getvalue()
        err_content = err_file.getvalue()
        # close buffers
        out_file.close()
        err_file.close()
        # create log files
        out_file, err_file = _create_out_and_err_files(export_path)
        # write buffered contents to log files
        out_file.write(out_content)
        err_file.write(err_content)
        out_file.close()
        err_file.close()
    else:
        if out_file != sys.stdout:
            out_file.close()
        if err_file != sys.stderr:
            err_file.close()


def _create_out_and_err_files(export_path):
    suffix = ''
    out_log_path = os.path.join(export_path, 'out.log')
    err_log_path = os.path.join(export_path, 'err.log')
    if os.path.exists(out_log_path) or os.path.exists(err_log_path):
        suffix += str(random.randint(constants.PRINT_SUFFIX_MIN, constants.PRINT_SUFFIX_MAX))
        out_log_path = os.path.join(export_path, f'out{suffix}.log')
        err_log_path = os.path.join(export_path, f'err{suffix}.log')
    out_file = open(out_log_path, 'w')
    err_file = open(err_log_path, 'w')

    return out_file, err_file


def crawl_website(
    export_path,
    base_url,
    options,
    action_type,
    scrape_option
):
    detection_strategy = OriginPageDetectionStrategy()
    if action_type == WebpageActionType.SCRAPE_PAGES:
        crawler = Crawler(options, None, export_path)
        for step in crawler.crawl(base_url):
            crawler.save(scrape_option)
    elif action_type == WebpageActionType.FIND_ORIGIN:
        crawler = Crawler(options, detection_strategy, export_path)
        for step in crawler.crawl(base_url):
            if detection_strategy.successful():
                break
    elif action_type == WebpageActionType.SCRAPE_PROFILES:
        crawler = Crawler(options, detection_strategy, export_path)
        for step in crawler.crawl(base_url):
            if detection_strategy.successful():
                result = detection_strategy.get_result()
                origin = result['origin']
                og_crawler = crawler.create_subcrawler()
                og_crawler.options.max_depth = result['depth'] + 1
                og_crawler.links_from_structure = True
                og_crawler.skip_sublinks = True
                og_crawler.skip_first_page = True
                og_crawler.sublink_filters = [lambda page_link: is_valid_sublink(page_link.url, result['most_common_format'], '####')]
                for og_step in og_crawler.crawl(origin, result['depth']):
                    og_crawler.save(scrape_option)
                    og_crawler.skip_sublinks = True
                crawler.mark_as_visited(og_crawler.get_visited_links(), og_crawler.get_scraped_count())


@dataclass
class CrawlOptions:
    max_depth: int = 3
    max_pages: int = None
    crawl_sleep: int = 2
    include_fragment: bool = False
    bump_relevant: bool = True
    use_buffer: bool = False
    scraping: bool = True
    resolution: tuple = (constants.WIDTH, constants.HEIGHT)
    classifier_name: object = None

    def increase(self, to_incr):
        for option, val in to_incr.items():
            assert val > 0, 'value must be greater then 0'
            if option == 'max_depth':
                self.max_depth += val
            elif option == 'max_pages':
                self.max_pages += val
            elif option == 'crawl_sleep':
                self.crawl_sleep += val
            else:
                raise KeyError(f'provided value {option!r} is not recognised as a crawl option')


class Crawler:
    def __init__(self, options, detection_strategy, export_path, parent_out_file=None, parent_err_file=None, is_subcrawler=False):
        self.skip_sublinks = False
        self.skip_first_page = False
        self.links_from_structure = False
        self.status = CrawlStatus.NOT_STARTED
        self.img_width = constants.WIDTH
        self.img_height = constants.HEIGHT
        self.sublink_filters = []
        self.detection_strategy = detection_strategy
        self.options = options
        self.is_subcrawler = is_subcrawler
        # prepare output files and directories
        self.export_path = export_path
        self.__out_file = sys.stdout
        self.__err_file = sys.stderr
        if is_subcrawler:
            self.__out_file = parent_out_file
            self.__err_file = parent_err_file
        else:
            export_path_exists = True
            try:
                os.mkdir(self.export_path)
            except FileExistsError:
                print(f'INFO: Directory exists at: {self.export_path}')
            except OSError:
                print(f'ERROR: Cannot create directory at {self.export_path!r}, stderr and stdout will be used')
                export_path_exists = False
            if export_path_exists:
                # open log files for writing if the directory is created
                if self.options.use_buffer:
                    self.__out_file = StringIO()
                    self.__err_file = StringIO()
                else:
                    self.__out_file, self.__err_file = _create_out_and_err_files(self.export_path)
            else:
                self.export_path = '.'
            if self.options.scraping:
                try:
                    os.mkdir(os.path.join(self.export_path, 'html'))
                    os.mkdir(os.path.join(self.export_path, 'screenshots'))
                except Exception:
                    pass

    def __visit_page(self):
        if not self.crawl_manager.has_next():
            print(f'INFO: All links at a depth of {self.options.max_depth} have been visited.',
                  'Stopping the crawling...',
                  file=self.__out_file)
            self.status = CrawlStatus.FINISHED
            return None
        self.curr_page = self.crawl_manager.visit_next()
        if self.curr_page is None:
            return None
        print(f'{self.curr_page.link.depth} {self.curr_page.link.url}', file=self.__out_file, flush=True)
        return self.curr_page

    def __visit_cleanup(self):
        self.__out_file.flush()
        self.__err_file.flush()
        time.sleep(self.options.crawl_sleep)

    def __perform_detection_strategy(self):
        self.detection_strategy.analyse(self.curr_page, self.options.classifier_name, self.options.resolution)
        result = self.detection_strategy.get_result()
        if result is not None:
            print(f"INFO: {result['message']}", file=self.__out_file)
        return result

    def __perform_action(self, action, args):
        action_result = action(**args)
        if action_result.successful:
            self.crawl_manager.increase_count()
            if self.crawl_manager.is_page_max_reached():
                self.status = CrawlStatus.FINISHED
        else:
            action_name = action.__name__
            print(f'ERROR: Failed to perform action {action_name!r} for: {self.curr_page.link.url}', file=self.__err_file)
        return action_result

    def __queue_sublinks(self):
        return self.crawl_manager.queue_sublinks(
            self.options.include_fragment,
            self.sublink_filters,
            self.links_from_structure)

    def crawl(self, base_url, base_depth=0):
        if not self.is_subcrawler:
            print(f'INFO: Logs for {base_url!r} are located at {self.export_path!r}')
        self.__web_driver = setup_web_driver()
        self.status = CrawlStatus.RUNNING
        self.crawl_manager = CrawlManager(self.__web_driver, base_url, self.__out_file, self.__err_file, base_depth=base_depth)
        self.crawl_manager.set_options(
            self.options.max_depth,
            self.options.max_pages,
            self.options.bump_relevant)
        try:
            while True:
                self.skip_sublinks = False
                current_page = self.__visit_page()
                if current_page is None or self.status == CrawlStatus.FINISHED:
                    break

                if self.detection_strategy is not None:
                    if self.detection_strategy.successful():
                        self.detection_strategy.reset()  # prep for new origin page
                    self.__perform_detection_strategy()

                if self.skip_first_page:
                    self.skip_first_page = False
                    print(f'INFO: Skipped page: {self.curr_page.link.url}', file=self.__out_file)
                else:
                    yield current_page.link

                # execution will resume here so check status to find out if crawling should be finished or not
                # note: caller might do something like performing action, which can lead to change of the
                #       crawl status
                if self.status == CrawlStatus.FINISHED:
                    break

                if not self.skip_sublinks:
                    self.__queue_sublinks()
                self.__visit_cleanup()
        except RemoteDisconnected as rde:
            print(f'INFO: Interrupted. Exiting... ({rde!r})', file=self.__err_file)
        except Exception as e:
            print(f'ERROR: {e!s}')
            print(f'{traceback.format_exc()}')
        finally:
            if not self.is_subcrawler:
                _close_everything(self.__web_driver, self.__out_file, self.__err_file, self.export_path, self.options.use_buffer)
                print(f'INFO: Crawling of {base_url!r} is complete')
            else:
                print(f'INFO: Subcrawling of {base_url!r} is complete')

    def create_subcrawler(self):
        options = copy.copy(self.options)
        return Crawler(options, None, self.export_path, self.__out_file, self.__err_file, True)

    def save(self, scrape_option):
        action = self.curr_page.scrape_page
        if scrape_option == ScrapeOption.ALL:
            args = {'export_path': self.export_path, 'scrape_option': ScrapeOption.ALL,
                    'width': self.img_width, 'height': self.img_height}
        elif scrape_option == ScrapeOption.HTML:
            action = self.curr_page.get_html
            args = {}
        elif scrape_option == ScrapeOption.SCREENSHOT:
            args = {'export_path': self.export_path, 'scrape_option': ScrapeOption.SCREENSHOT,
                    'width': self.img_width, 'height': self.img_height}
        else:
            return None
        return self.__perform_action(action, args)

    def get_visited_links(self):
        return self.crawl_manager.get_visited_links()

    def get_links_to_visit(self):
        return self.crawl_manager.get_links_to_visit()

    def get_scraped_count(self):
        return self.crawl_manager.get_scraped_count()

    def mark_as_visited(self, visited_links, scraped_count):
        return self.crawl_manager.mark_as_visited(visited_links, scraped_count)



class ConstantsNamespace:
    __slots__ = ()

    # File extension of the screenshot
    IMG_EXT = 'png'

    # Default image resolution
    WIDTH = 2880
    HEIGHT = 1620

    # Time for web driver to wait implicitly before attempting to
    # perform the action
    IMPL_WAIT_FOR_FULL_LOAD = 18

    # Time to wait before next attempt to interact with web driver is made.
    # Used for page visit
    RETRY_TIME = 60

    # Threshold for using buffer instead of printing everything to the file
    # line by line
    BUFF_THRESHOLD = 30

    # List of file extensions to ignore if a resource with that
    # extension is visited
    INVALID_EXTENSIONS = [
        'mp4', 'jpg', 'png', 'jpeg',
        'zip', 'rar', 'xls', 'rtf',
        'docx', 'doc', 'pptx', 'ppt',
        'pdf', 'txt']

    # List of words that are relevant for the profile.
    # Program will match them with the URL and possibly prioritize
    # if requested
    RELEVANT_WORDS = [
        # en
        'profile', 'user', 'users',
        'about-us', 'team',
        'employees', 'staff', 'professor',

        # rs
        'profil',
        'o-nama',
        'zaposlen', 'nastavnik', 'nastavnici', 'saradnici', 'profesor', 'osoblje',
        'запослен', 'наставник', 'наставници', 'сарадници', 'професор', 'особље']

    # Suffix that is added to filename if file with default filename is present
    PRINT_SUFFIX_MIN = 100_000
    PRINT_SUFFIX_MAX = 999_999

    # Threshold value for detecting positive class
    PREDICTION_THRESHOLD = 0.5

    # The threshold determines the number of subpages of a URL that are categorized as profile pages.
    # Once this threshold is met, the URL is regarded as the parent page for all subsequent profile pages
    ORIGIN_PAGE_THRESHOLD = 3

    IMAGE_CLASSIFIERS = [
        'scooby',
        'batman']

    # Max lenght of the filename. This is desired limit. If target filesystem does not support
    # this lenght then it's limit will be used as maximum lenght
    # Note: Kaggle has limit of 99 characters for filename
    FILENAME_MAX_LENGHT = 99

    # Suffix that will be appended after cut if filename is too long.
    # Used to indicate that at that point where cropping happened.
    # There is also another part of the suffix which is appended to avoid same filename conflicts
    FILENAME_CUT_SUFFIX = '--CROP_'

    # Mapping for unsafe chars that may appear in the filename
    CHAR_REPLACEMENTS = {
                        '#': 'ANCH',
                        '?': 'QMARK',
                        '&': 'AMP',
                        '@': 'ATSGN',
                        '!': 'EMARK',
                        ':': 'COL',
                        ';': 'SEMICOL',
                        ',': 'COMMA',
                        "'": 'APOST',
                        '"': 'QUOTE',
                        '`': 'BTICK',
                        '(': 'BR',
                        ')': 'BR',
                        '{': 'CRBR',
                        '}': 'CRBR',
                        '[': 'SQBR',
                        ']': 'SQBR',
                        '<': 'LTHEN',
                        '>': 'GTHEN',
                        '/': '__',
                        '|': 'PIPE',
                        '\\': 'BSLASH',
                        '%': 'PERC',
                        '+': 'PLUS',
                        '*': 'STAR',
                        '=': 'EQL',
                        '^': 'CARET',
                        '~': 'TILDA'}



from enum import Enum

from profilescout.web.webpage import Webpage
from profilescout.common.constants import ConstantsNamespace
from profilescout.common.exceptions import WebDriverException, parse_web_driver_exception
from profilescout.link.utils import PageLink, remove_duplicates, prioritize_relevant, to_abs_path
from profilescout.link.utils import filter_out_invalid, filter_out_visited, filter_out_present_links, filter_out_long


constants = ConstantsNamespace

CrawlStatus = Enum('CrawlStatus', ['NOT_STARTED', 'RUNNING', 'FINISHED'])


class CrawlManager:
    def __init__(self, web_driver, base_url, out_file, err_file, max_depth=3, max_pages=None, base_depth=0):
        self.__web_driver = web_driver
        self.__base_url = base_url
        self.__max_depth = max_depth
        self.__max_pages = max_pages
        self.__bump_relevant = False

        page_link = PageLink(base_url, base_depth)
        self.curr_page = Webpage(web_driver, page_link, out_file, err_file)

        self.__scraped_count = 0
        self.__visited_links = set()
        self.__links_to_visit = [page_link]  # add base url as link that needs to be visited

        self.__out_file = out_file
        self.__err_file = err_file

    def __set_curr_page(self, page_link):
        self.curr_page = Webpage(self.__web_driver, page_link, self.__out_file, self.__err_file)
        return self.curr_page

    def has_next(self):
        return len(self.__links_to_visit) > 0

    def clear_history(self, init_page=None):
        self.__scraped_count = 0
        self.__visited_links = set()

        link = init_page if init_page is not None else self.curr_page
        self.__links_to_visit = [link]

    def mark_as_visited(self, urls, scraped_count):
        self.__scraped_count += scraped_count
        self.__visited_links.update(urls)
        self.__links_to_visit = [pl for pl in self.__links_to_visit if pl.url not in urls]
        return self.__links_to_visit, self.__visited_links

    def get_visited_links(self):
        return self.__visited_links

    def get_links_to_visit(self):
        return self.__links_to_visit

    def get_scraped_count(self):
        return self.__scraped_count

    def set_options(self, max_depth=None, max_pages=None, bump_relevant=None):
        self.__max_depth = max_depth if max_depth is not None else self.__max_depth
        self.__max_pages = max_pages if max_pages is not None else self.__max_pages
        self.__bump_relevant = bump_relevant if bump_relevant is not None else self.__bump_relevant

    def is_page_max_reached(self):
        if self.__max_pages is not None and self.__scraped_count == self.__max_pages:
            print(f'INFO: Maximum number of pages to scrape ({self.__max_pages}) reached.',
                  'Stopping the crawling...',
                  file=self.__out_file)
            print(f'INFO: There were {len(self.__links_to_visit)} unvisited links in the queue',
                  file=self.__out_file)
            print(f'INFO: Current depth: {self.curr_page.link.depth}', file=self.__out_file)
            return True
        return False

    def visit_next(self):
        # take next link from the queue and visit it
        self.__set_curr_page(self.__links_to_visit.pop(0))

        # visit page
        try:
            is_text_file = self.curr_page.visit()
        except WebDriverException as e:
            err_msg, reason = parse_web_driver_exception(e, self.curr_page.link.url)
            print(f'ERROR: {err_msg} (reason: {reason})', file=self.__err_file)
            print(f'WARN: {reason} {self.curr_page.link.url}', file=self.__out_file)
            # skip this url
            return None

        # mark link as visited
        self.__visited_links.add(self.curr_page.link.url)

        # if content-type is not 'text/*' then ignore it
        if not is_text_file:
            return None
        return self.curr_page

    def increase_count(self):
        self.__scraped_count += 1
        return self.__scraped_count

    def get_options(self):
        return {
            'max_depth': self.__max_depth,
            'max_pages': self.__max_pages,
            'bump_relevant': self.__bump_relevant}

    def queue_sublinks(self, include_fragment=False, link_filters=[], links_from_structure=False):
        # check if the maximum depth (number of hops) has been reached
        if self.curr_page.link.depth == self.__max_depth:
            # ignore links on the current page and continue with visiting links that left to be visited
            return None
        # extract URLs
        if not links_from_structure or not hasattr(self, '_CrawlManager__previous_links'):
            self.__previous_links = []
        hops = self.curr_page.extract_links(self.__base_url, include_fragment, links_from_structure,  self.__previous_links)
        self.__previous_links = [hop.url for hop in hops]

        # transform extracted URLs, as some of them may be invalid or irrelevant
        hops_with_abs_path = [PageLink(
                to_abs_path(pl.url, self.curr_page.link.url),
                pl.depth,
                pl.parent_url,
                pl.txt)
            for pl in hops]
        valid = filter_out_invalid(hops_with_abs_path, self.__base_url)
        valid_not_visited = filter_out_visited(valid, self.__visited_links)
        new_links = filter_out_present_links(valid_not_visited, self.__links_to_visit)
        new_links = filter_out_long(new_links, self.__err_file)
        new_links = remove_duplicates(new_links)
        if link_filters != []:
            for link_filter in link_filters:
                new_links = filter(link_filter, new_links)
        self.__links_to_visit.extend(new_links)

        if self.__bump_relevant:
            self.__links_to_visit = prioritize_relevant(self.__links_to_visit)

        return self.__links_to_visit



import platform
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.common.by import By
from selenium.common.exceptions import (
    StaleElementReferenceException as SeleniumStaleElementReferenceException,
    WebDriverException as SeleniumWebDriverException)

from profilescout.common.constants import ConstantsNamespace
from profilescout.common.wrappers import WebElementWrapper, WebDriverWrapper
from profilescout.common.exceptions import StaleElementReferenceException, WebDriverException


constants = ConstantsNamespace


def setup_web_driver():
    # create a new Chrome browser instance Options
    options = webdriver.ChromeOptions()

    # disable file downloads
    null_path = '/dev/null'  # assume that program is run on Unix-like OS

    if platform.system() == 'Windows':
        null_path = 'NUL'

    profile = {
            "plugins.plugins_list": [{"enabled": False, "name": "Chrome PDF Viewer"}],
            "download.default_directory": null_path,
            "profile.default_content_settings.popups": 0,
            "download.prompt_for_download": False,
            "download_restrictions": 3,  # https://chromeenterprise.google/policies/#DownloadRestrictions
            "safebrowsing.enabled": True
        }

    options.add_experimental_option("prefs", profile)

    options.add_argument("--no-sandbox")                 # bypass OS security model
    options.add_argument("--start-maximized")            # open Browser in maximized mode
    options.add_argument("--disable-extensions")         # disabling extensions
    options.add_argument("--disable-gpu")                # applicable to Windows only
    options.add_argument("--disable-dev-shm-usage")      # overcome limited resource problems
    options.add_argument("--disable-application-cache")
    options.add_argument("--mute-audio")

    # disable GUI
    options.add_argument('--headless')
    options.add_argument('--disable-infobars')

    web_driver = webdriver.Chrome(
        service=Service(),
        options=options)

    # wait for the page to fully load
    web_driver.implicitly_wait(constants.IMPL_WAIT_FOR_FULL_LOAD)

    return WebDriver(web_driver)


class WebElement(WebElementWrapper):
    """Implementation of WebElementWrapper that wraps WebElement."""

    def __init__(self, element):
        self.element = element

    def get_attribute(self, name):
        try:
            return self.element.get_attribute(name)
        except SeleniumStaleElementReferenceException as e:
            raise StaleElementReferenceException.from_stale_element_exception(e)

    def find_elements_with_xpath(self, xpath):
        return [WebElement(el) for el in self.element.find_elements(By.XPATH, xpath)]

    @property
    def text(self):
        return self.element.text


class WebDriver(WebDriverWrapper):
    """Implementation of WebDriverWrapper that wraps WebDriver."""

    def __init__(self, driver):
        self.__driver = driver

    def get(self, url):
        try:
            return self.__driver.get(url)
        except SeleniumWebDriverException as e:
            raise WebDriverException.from_webdriver_exception(e)

    def get_screenshot_as_png(self):
        return self.__driver.get_screenshot_as_png()

    def save_screenshot(self, path):
        return self.__driver.save_screenshot(path)

    def get_page_source(self):
        return self.__driver.page_source

    def find_elements_with_xpath(self, xpath):
        return [WebElement(el) for el in self.__driver.find_elements(By.XPATH, xpath)]

    def execute_script(self, script):
        return self.__driver.execute_script(script)

    def set_window_size(self, width, height):
        return self.__driver.set_window_size(width, height)

    def quit(self):
        self.__driver.quit()



from abc import ABC, abstractmethod


class WebElementWrapper(ABC):
    """Interface/Base class to abstract WebElement usage."""

    @abstractmethod
    def get_attribute(self, name):
        pass

    @abstractmethod
    def find_elements_with_xpath(self, xpath):
        pass


class WebDriverWrapper(ABC):
    '''
    Interface/Base class to abstract WebDriver usage.
    '''

    @abstractmethod
    def get(self, url):
        pass

    @abstractmethod
    def get_screenshot_as_png(self):
        pass

    @abstractmethod
    def save_screenshot(self, path):
        pass

    @abstractmethod
    def get_page_source(self):
        pass

    @abstractmethod
    def find_elements_with_xpath(self, xpath):
        pass

    @abstractmethod
    def execute_script(self, script):
        pass

    @abstractmethod
    def set_window_size(self, width, height):
        pass

    @abstractmethod
    def quit(self):
        pass
